
from anynet import tls, http
import json

import logging
logger = logging.getLogger(__name__)


MODULE_ACCOUNT = "nnAccount"
MODULE_FRIENDS = "nnFriends"

USER_AGENT = {
	 900: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 901: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 910: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	 920: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 9.3.0.0; Add-on 9.3.0.0)",
	1000: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1001: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1002: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1003: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1004: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1010: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1011: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1020: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 10.4.0.0; Add-on 10.4.0.0)",
	1100: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1101: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 11.4.0.0; Add-on 11.4.0.0)",
	1200: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1201: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1202: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1203: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1210: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 12.3.0.0; Add-on 12.3.0.0)",
	1300: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.3.0.0; Add-on 13.3.0.0)",
	1310: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1320: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1321: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 13.4.0.0; Add-on 13.4.0.0)",
	1400: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1410: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1411: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1412: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 14.3.0.0; Add-on 14.3.0.0)",
	1500: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 15.3.0.0; Add-on 15.3.0.0)",
	1501: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 15.3.0.0; Add-on 15.3.0.0)",
	1600: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1601: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1602: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1603: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1610: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 16.2.0.0; Add-on 16.2.0.0)",
	1700: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 17.5.0.0; Add-on 17.5.0.0)",
	1701: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 17.5.0.0; Add-on 17.5.0.0)",
	1800: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1801: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1810: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 18.3.0.0; Add-on 18.3.0.0)",
	1900: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 19.3.0.0; Add-on 19.3.0.0)",
	1901: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 19.3.0.0; Add-on 19.3.0.0)",
	2000: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2001: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2010: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2011: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2015: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2020: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2030: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2040: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2050: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 20.5.4.0; Add-on 20.5.4.0)",
	2100: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
	2101: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
	2110: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
	2120: "libcurl (%s; 789f928b-138e-4b2f-afeb-1acae821d897; SDK 21.4.0.0; Add-on 21.4.0.0)",
}

LATEST_VERSION = 2120


class PresenceState:
	INACTIVE = "INACTIVE"
	ONLINE = "ONLINE"
	PLAYING = "PLAYING"


class BAASError(Exception):
	def __init__(self, response):
		self.response = response
		
		self.type = response.json["type"]
		self.name = response.json["errorCode"]
		self.title = response.json["title"]
		self.detail = response.json["detail"]
		self.status = response.json["status"]
		self.instance = response.json["instance"]
	
	def __str__(self):
		return self.title


class BAASClient:
	def __init__(self):
		self.request_callback = http.request
		
		self.context = tls.TLSContext()
		
		self.host = None
		self.power_state = "FA"

		self.system_version = LATEST_VERSION
		self.user_agent = USER_AGENT[LATEST_VERSION]
	
	def set_request_callback(self, callback): self.request_callback = callback
	def set_context(self, context): self.context = context
	
	def set_certificate(self, cert, key):
		self.context.set_certificate(cert, key)
	
	def set_host(self, host): self.host = host
	def set_power_state(self, state): self.power_state = state
	
	def set_system_version(self, version):
		if version not in USER_AGENT:
			raise ValueError("Unknown system version")
		self.system_version = version
		self.user_agent = USER_AGENT[version]
	
	def get_host(self, module):
		if self.host:
			return self.host
		
		if module == MODULE_ACCOUNT and self.system_version >= 2100:
			return "m-lp1.baas.nintendo.com"
		return "e0d67c509fb203858ebcb2fe3f88c2aa.baas.nintendo.com"
		
	async def request(self, req, token, module, *, use_power_state=False):
		# This is somewhat complicated because we want to
		# put the headers in the correct order
		content_type = "application/x-www-form-urlencoded"
		if req.json is not None:
			content_type = "application/json-patch+json" if req.method == "PATCH" else "application/json"
		
		req.headers["Host"] = self.host
		if module == MODULE_ACCOUNT and self.system_version >= 2100:
			req.headers["Accept"] = "*/*"
			req.headers["User-Agent"] = self.user_agent %module
			if req.method != "GET":
				req.headers["Content-Type"] = content_type
			if token:
				req.headers["Authorization"] = "Bearer " + token
			if use_power_state:
				req.headers["X-Nintendo-PowerState"] = self.power_state
			if req.method != "GET":
				req.headers["Content-Length"] = 0
		else:
			req.headers["User-Agent"] = self.user_agent %module
			req.headers["Accept"] = "*/*"
			if module == MODULE_FRIENDS and req.json is not None:
				req.headers["Content-Type"] = content_type
			if token:
				req.headers["Authorization"] = "Bearer " + token
			if use_power_state:
				req.headers["X-Nintendo-PowerState"] = self.power_state
			
			if req.method != "GET":
				if req.json is None:
					req.headers["Content-Length"] = 0
					req.headers["Content-Type"] = content_type
				else:
					if module != MODULE_FRIENDS:
						req.headers["Content-Type"] = content_type
					req.headers["Content-Length"] = 0
		
		response = await self.request_callback(self.host, req, self.context)
		if response.json and "errorCode" in response.json:
			logger.warning("BAAS server returned an error: %s" %response.json)
			raise BAASError(response)
		response.raise_if_error()
		return response
	
	async def authenticate(self, device_token, penne_id=None):
		req = http.HTTPRequest.post("/1.0.0/application/token")
		req.form = {
			"grantType": "public_client",
			"assertion": device_token
		}
		if self.system_version >= 1900 and penne_id is not None:
			req.form["penneId"] = penne_id

		response = await self.request(req, None, MODULE_ACCOUNT, use_power_state=True)
		return response.json
	
	async def login(self, id, password, access_token, app_token=None, na_country=None, skip_verification=False, is_persistent=True):
		req = http.HTTPRequest.post("/1.0.0/login")
		req.form = {
			"id": "%016x" %id,
			"password": password
		}
		
		if app_token:
			req.form["appAuthNToken"] = app_token
		
		if self.system_version >= 1800:
			if na_country is None:
				raise ValueError("na_country parameter is required for system version 18.0.0 and later")
			req.form["naCountry"] = na_country
		
		if self.system_version >= 2000:
			req.form["isPersistent"] = "true" if is_persistent else "false"
		
		if skip_verification:
			req.form["skipOp2Verification"] = "1"
			
		response = await self.request(req, access_token, MODULE_ACCOUNT, use_power_state=True)
		return response.json
	
	async def register(self, access_token):
		req = http.HTTPRequest.post("/1.0.0/users")
		
		response = await self.request(req, access_token, MODULE_ACCOUNT)
		return response.json
	
	async def update_presence(self, user_id, device_account_id, access_token, state, title_id, presence_group_id, app_fields={}, acd_index=0):
		app_fields = json.dumps(app_fields, separators=(",", ":"))
		
		req = http.HTTPRequest.patch("/1.0.0/users/%016x/device_accounts/%016x" %(user_id, device_account_id))
		req.json = [
			{"op": "replace", "path": "/presence/state", "value": state},
			{"op": "add", "path": "/presence/extras/friends/appField", "value": app_fields},
			{"op": "add", "path": "/presence/extras/friends/appInfo:appId", "value": "%016x" %title_id}
		]
		if self.system_version >= 1900:
			req.json.append({"op": "add", "path": "/presence/extras/friends/appInfo:acdIndex", "value": acd_index})
		req.json.append({"op": "add", "path": "/presence/extras/friends/appInfo:presenceGroupId", "value": "%016x" %presence_group_id})

		response = await self.request(req, access_token, MODULE_FRIENDS)
		return response.json
	
	async def get_friends(self, user_id, access_token, count=300):
		req = http.HTTPRequest.get("/2.0.0/users/%016x/friends" %user_id)
		req.params = {"count": count}
		response = await self.request(req, access_token, MODULE_FRIENDS)
		return response.json




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class ScreeningProtocol:
	METHOD_REPORT_DATA_STORE_CONTENT = 1
	METHOD_REPORT_USER = 2
	
	PROTOCOL_ID = 0x7C


class ScreeningClient(ScreeningProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	


class ScreeningServer(ScreeningProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_REPORT_DATA_STORE_CONTENT: self.handle_report_data_store_content,
			self.METHOD_REPORT_USER: self.handle_report_user,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on ScreeningServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_report_data_store_content(self, client, input, output):
		logger.warning("ScreeningServer.report_data_store_content is not supported")
		raise common.RMCError("Core::NotImplemented")
	
	async def handle_report_user(self, client, input, output):
		logger.warning("ScreeningServer.report_user is not supported")
		raise common.RMCError("Core::NotImplemented")





from anynet import tls
import importlib.resources


def get(path):
	return importlib.resources.files("nintendo").joinpath(path)

def open(path, mode="r", *, encoding="utf-8"):
	return get(path).open(mode, encoding=encoding)

def read_bytes(path):
	return get(path).read_bytes()

def certificate(name):
	data = read_bytes("files/cert/%s" %name)
	return tls.TLSCertificate.parse(data, tls.TYPE_DER)

def private_key(name):
	data = read_bytes("files/cert/%s" %name)
	return tls.TLSPrivateKey.parse(data, tls.TYPE_DER)




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class ApplicationInfo(common.Structure):
	def __init__(self):
		super().__init__()
		self.title_id = None
		self.title_version = None
	
	def check_required(self, settings, version):
		for field in ['title_id', 'title_version']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.title_id = stream.u64()
		self.title_version = stream.u16()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u64(self.title_id)
		stream.u16(self.title_version)


class AAUserProtocol:
	METHOD_REGISTER_APPLICATION = 1
	METHOD_UNREGISTER_APPLICATION = 2
	METHOD_SET_APPLICATION_INFO = 3
	METHOD_GET_APPLICATION_INFO = 4
	
	PROTOCOL_ID = 0x7B


class AAUserClient(AAUserProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def register_application(self, title_id):
		logger.info("AAUserClient.register_application()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(title_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REGISTER_APPLICATION, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AAUserClient.register_application -> done")
	
	async def unregister_application(self, title_id):
		logger.info("AAUserClient.unregister_application()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u64(title_id)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UNREGISTER_APPLICATION, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AAUserClient.unregister_application -> done")
	
	async def set_application_info(self, application_info):
		logger.info("AAUserClient.set_application_info()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(application_info, stream.add)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_SET_APPLICATION_INFO, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AAUserClient.set_application_info -> done")
	
	async def get_application_info(self):
		logger.info("AAUserClient.get_application_info()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_APPLICATION_INFO, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		info = stream.list(ApplicationInfo)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AAUserClient.get_application_info -> done")
		return info


class AAUserServer(AAUserProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_REGISTER_APPLICATION: self.handle_register_application,
			self.METHOD_UNREGISTER_APPLICATION: self.handle_unregister_application,
			self.METHOD_SET_APPLICATION_INFO: self.handle_set_application_info,
			self.METHOD_GET_APPLICATION_INFO: self.handle_get_application_info,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on AAUserServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_register_application(self, client, input, output):
		logger.info("AAUserServer.register_application()")
		#--- request ---
		title_id = input.u64()
		await self.register_application(client, title_id)
	
	async def handle_unregister_application(self, client, input, output):
		logger.info("AAUserServer.unregister_application()")
		#--- request ---
		title_id = input.u64()
		await self.unregister_application(client, title_id)
	
	async def handle_set_application_info(self, client, input, output):
		logger.info("AAUserServer.set_application_info()")
		#--- request ---
		application_info = input.list(ApplicationInfo)
		await self.set_application_info(client, application_info)
	
	async def handle_get_application_info(self, client, input, output):
		logger.info("AAUserServer.get_application_info()")
		#--- request ---
		response = await self.get_application_info(client)
		
		#--- response ---
		if not isinstance(response, list):
			raise RuntimeError("Expected list, got %s" %response.__class__.__name__)
		output.list(response, output.add)
	
	async def register_application(self, *args):
		logger.warning("AAUserServer.register_application not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def unregister_application(self, *args):
		logger.warning("AAUserServer.unregister_application not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def set_application_info(self, *args):
		logger.warning("AAUserServer.set_application_info not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_application_info(self, *args):
		logger.warning("AAUserServer.get_application_info not implemented")
		raise common.RMCError("Core::NotImplemented")





from Crypto.Cipher import AES
from Crypto.PublicKey import RSA
from anynet import tls
import hashlib
import struct


def load_keys(filename):
	with open(filename) as f:
		lines = f.readlines()
	
	keys = {}
	for line in lines:
		line = line.strip()
		if line:
			name, key = line.split("=")
			keys[name.strip()] = bytes.fromhex(key)
	return keys


table = [
	0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401,
	0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400
]

def crc16(data):
	hash = 0x55AA
	for byte in data:
		r = table[hash & 0xF]
		hash = (hash >> 4) ^ r ^ table[byte & 0xF]
		
		r = table[hash & 0xF]
		hash = (hash >> 4) ^ r ^ table[byte >> 4]
	return hash


class ProdInfo:
	def __init__(self, keys, filename):
		self.keys = keys
		with open(filename, "rb") as f:
			self.data = f.read()
			
	def check(self, offset, size):
		end = offset + size - 2
		
		expected = struct.unpack_from("<H", self.data, end)[0]
		if crc16(self.data[offset : end]) != expected:
			raise ValueError("CRC16 check failed")
	
	def get_device_id(self):
		self.check(0x2A90, 0x250)
		return int(self.data[0x2B56:0x2B66], 16)
	
	def get_tls_cert(self):
		self.check(0xAD0, 0x10)
		
		length = struct.unpack_from("<I", self.data, 0xAD0)[0]
		if length > 0x800:
			raise ValueError("TLS certificate is too big")
			
		data = self.data[0xAE0 : 0xAE0 + length]
		hash = hashlib.sha256(data).digest()
		if hash != self.data[0x12E0 : 0x1300]:
			raise ValueError("SHA256 check failed")
		
		return tls.TLSCertificate.parse(data, tls.TYPE_DER)
		
	def get_tls_key(self):
		self.check(0x3AE0, 0x140)
		
		initial = self.data[0x3AE0 : 0x3AF0]
		cipher = self.data[0x3AF0 : 0x3BF0]
		
		kek = self.keys["ssl_rsa_kek_personalized"] if "ssl_rsa_kek_personalized" in self.keys else self.keys["ssl_rsa_kek"]
		aes = AES.new(kek, AES.MODE_CTR, nonce=b"", initial_value=initial)
		d = int.from_bytes(aes.decrypt(cipher), "big")
		
		pubkey = self.get_tls_cert().public_key()
		
		rsa = RSA.construct((pubkey.n, pubkey.e, d))
		der = rsa.export_key("DER")
		return tls.TLSPrivateKey.parse(der, tls.TYPE_DER)




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class AuthenticationInfo(common.Data):
	def __init__(self):
		super().__init__()
		self.token = None
		self.ngs_version = 3
		self.token_type = 1
		self.server_version = 0
	
	def check_required(self, settings, version):
		for field in ['token']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.token = stream.string()
		self.ngs_version = stream.u32()
		self.token_type = stream.u8()
		self.server_version = stream.u32()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.string(self.token)
		stream.u32(self.ngs_version)
		stream.u8(self.token_type)
		stream.u32(self.server_version)
common.DataHolder.register(AuthenticationInfo, "AuthenticationInfo")


class RVConnectionData(common.Structure):
	def __init__(self):
		super().__init__()
		self.main_station = common.StationURL.parse("prudp:/")
		self.special_protocols = []
		self.special_station = common.StationURL.parse("prudp:/")
		self.server_time = common.DateTime(0)
	
	def max_version(self, settings):
		version = 0
		if settings["nex.version"] >= 30500:
			version = 1
		return version
	
	def check_required(self, settings, version):
		if settings["nex.version"] >= 30500:
			if version >= 1:
				pass
	
	def load(self, stream, version):
		self.main_station = stream.stationurl()
		self.special_protocols = stream.list(stream.u8)
		self.special_station = stream.stationurl()
		if stream.settings["nex.version"] >= 30500:
			if version >= 1:
				self.server_time = stream.datetime()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.stationurl(self.main_station)
		stream.list(self.special_protocols, stream.u8)
		stream.stationurl(self.special_station)
		if stream.settings["nex.version"] >= 30500:
			if version >= 1:
				stream.datetime(self.server_time)


class ValidateAndRequestTicketParam(common.Structure):
	def __init__(self):
		super().__init__()
		self.platform = 3
		self.username = None
		self.data = None
		self.skip_version_check = False
		self.nex_version = None
		self.client_version = None
	
	def check_required(self, settings, version):
		for field in ['username', 'data', 'nex_version', 'client_version']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.platform = stream.u32()
		self.username = stream.string()
		self.data = stream.anydata()
		self.skip_version_check = stream.bool()
		self.nex_version = stream.u32()
		self.client_version = stream.u32()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.u32(self.platform)
		stream.string(self.username)
		stream.anydata(self.data)
		stream.bool(self.skip_version_check)
		stream.u32(self.nex_version)
		stream.u32(self.client_version)


class ValidateAndRequestTicketResult(common.Structure):
	def __init__(self):
		super().__init__()
		self.pid = None
		self.ticket = None
		self.server_url = None
		self.server_time = None
		self.server_name = None
		self.source_key = None
	
	def check_required(self, settings, version):
		for field in ['pid', 'ticket', 'server_url', 'server_time', 'server_name', 'source_key']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.pid = stream.pid()
		self.ticket = stream.buffer()
		self.server_url = stream.stationurl()
		self.server_time = stream.datetime()
		self.server_name = stream.string()
		self.source_key = stream.string()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.pid(self.pid)
		stream.buffer(self.ticket)
		stream.stationurl(self.server_url)
		stream.datetime(self.server_time)
		stream.string(self.server_name)
		stream.string(self.source_key)


class AuthenticationProtocol:
	METHOD_LOGIN = 1
	METHOD_LOGIN_EX = 2
	METHOD_REQUEST_TICKET = 3
	METHOD_GET_PID = 4
	METHOD_GET_NAME = 5
	METHOD_LOGIN_WITH_CONTEXT = 6
	
	PROTOCOL_ID = 0xA


class AuthenticationProtocolNX:
	METHOD_VALIDATE_AND_REQUEST_TICKET = 1
	METHOD_VALIDATE_AND_REQUEST_TICKET_WITH_CUSTOM_DATA = 2
	METHOD_REQUEST_TICKET = 3
	METHOD_GET_PID = 4
	METHOD_GET_NAME = 5
	METHOD_VALIDATE_AND_REQUEST_TICKET_WITH_PARAM = 6
	
	PROTOCOL_ID = 0xA


class AuthenticationClient(AuthenticationProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def login(self, username):
		logger.info("AuthenticationClient.login()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(username)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_LOGIN, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.pid = stream.pid()
		obj.ticket = stream.buffer()
		obj.connection_data = stream.extract(RVConnectionData)
		obj.server_name = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClient.login -> done")
		return obj
	
	async def login_ex(self, username, extra_data):
		logger.info("AuthenticationClient.login_ex()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(username)
		stream.anydata(extra_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_LOGIN_EX, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.pid = stream.pid()
		obj.ticket = stream.buffer()
		obj.connection_data = stream.extract(RVConnectionData)
		obj.server_name = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClient.login_ex -> done")
		return obj
	
	async def request_ticket(self, source, target):
		logger.info("AuthenticationClient.request_ticket()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(source)
		stream.pid(target)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REQUEST_TICKET, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.ticket = stream.buffer()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClient.request_ticket -> done")
		return obj
	
	async def get_pid(self, username):
		logger.info("AuthenticationClient.get_pid()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(username)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_PID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		pid = stream.pid()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClient.get_pid -> done")
		return pid
	
	async def get_name(self, pid):
		logger.info("AuthenticationClient.get_name()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_NAME, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		name = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClient.get_name -> done")
		return name
	
	async def login_with_context(self, login_data):
		logger.info("AuthenticationClient.login_with_context()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.anydata(login_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_LOGIN_WITH_CONTEXT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.pid = stream.pid()
		obj.ticket = stream.buffer()
		obj.connection_data = stream.extract(RVConnectionData)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClient.login_with_context -> done")
		return obj


class AuthenticationClientNX(AuthenticationProtocolNX):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def validate_and_request_ticket(self, username):
		logger.info("AuthenticationClientNX.validate_and_request_ticket()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(username)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_VALIDATE_AND_REQUEST_TICKET, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.pid = stream.pid()
		obj.ticket = stream.buffer()
		obj.connection_data = stream.extract(RVConnectionData)
		obj.server_name = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClientNX.validate_and_request_ticket -> done")
		return obj
	
	async def validate_and_request_ticket_with_custom_data(self, username, extra_data):
		logger.info("AuthenticationClientNX.validate_and_request_ticket_with_custom_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(username)
		stream.anydata(extra_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_VALIDATE_AND_REQUEST_TICKET_WITH_CUSTOM_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.pid = stream.pid()
		obj.ticket = stream.buffer()
		obj.connection_data = stream.extract(RVConnectionData)
		obj.server_name = stream.string()
		obj.source_key = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClientNX.validate_and_request_ticket_with_custom_data -> done")
		return obj
	
	async def request_ticket(self, source, target):
		logger.info("AuthenticationClientNX.request_ticket()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(source)
		stream.pid(target)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REQUEST_TICKET, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.ticket = stream.buffer()
		obj.key = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClientNX.request_ticket -> done")
		return obj
	
	async def get_pid(self, username):
		logger.info("AuthenticationClientNX.get_pid()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.string(username)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_PID, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		pid = stream.pid()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClientNX.get_pid -> done")
		return pid
	
	async def get_name(self, pid):
		logger.info("AuthenticationClientNX.get_name()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_GET_NAME, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		name = stream.string()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClientNX.get_name -> done")
		return name
	
	async def validate_and_request_ticket_with_param(self, param):
		logger.info("AuthenticationClientNX.validate_and_request_ticket_with_param()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.add(param)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_VALIDATE_AND_REQUEST_TICKET_WITH_PARAM, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		result = stream.extract(ValidateAndRequestTicketResult)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("AuthenticationClientNX.validate_and_request_ticket_with_param -> done")
		return result


class AuthenticationServer(AuthenticationProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_LOGIN: self.handle_login,
			self.METHOD_LOGIN_EX: self.handle_login_ex,
			self.METHOD_REQUEST_TICKET: self.handle_request_ticket,
			self.METHOD_GET_PID: self.handle_get_pid,
			self.METHOD_GET_NAME: self.handle_get_name,
			self.METHOD_LOGIN_WITH_CONTEXT: self.handle_login_with_context,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on AuthenticationServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_login(self, client, input, output):
		logger.info("AuthenticationServer.login()")
		#--- request ---
		username = input.string()
		response = await self.login(client, username)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'pid', 'ticket', 'connection_data', 'server_name']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.pid(response.pid)
		output.buffer(response.ticket)
		output.add(response.connection_data)
		output.string(response.server_name)
	
	async def handle_login_ex(self, client, input, output):
		logger.info("AuthenticationServer.login_ex()")
		#--- request ---
		username = input.string()
		extra_data = input.anydata()
		response = await self.login_ex(client, username, extra_data)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'pid', 'ticket', 'connection_data', 'server_name']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.pid(response.pid)
		output.buffer(response.ticket)
		output.add(response.connection_data)
		output.string(response.server_name)
	
	async def handle_request_ticket(self, client, input, output):
		logger.info("AuthenticationServer.request_ticket()")
		#--- request ---
		source = input.pid()
		target = input.pid()
		response = await self.request_ticket(client, source, target)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'ticket']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.buffer(response.ticket)
	
	async def handle_get_pid(self, client, input, output):
		logger.info("AuthenticationServer.get_pid()")
		#--- request ---
		username = input.string()
		response = await self.get_pid(client, username)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.pid(response)
	
	async def handle_get_name(self, client, input, output):
		logger.info("AuthenticationServer.get_name()")
		#--- request ---
		pid = input.pid()
		response = await self.get_name(client, pid)
		
		#--- response ---
		if not isinstance(response, str):
			raise RuntimeError("Expected str, got %s" %response.__class__.__name__)
		output.string(response)
	
	async def handle_login_with_context(self, client, input, output):
		logger.info("AuthenticationServer.login_with_context()")
		#--- request ---
		login_data = input.anydata()
		response = await self.login_with_context(client, login_data)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'pid', 'ticket', 'connection_data']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.pid(response.pid)
		output.buffer(response.ticket)
		output.add(response.connection_data)
	
	async def login(self, *args):
		logger.warning("AuthenticationServer.login not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def login_ex(self, *args):
		logger.warning("AuthenticationServer.login_ex not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def request_ticket(self, *args):
		logger.warning("AuthenticationServer.request_ticket not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_pid(self, *args):
		logger.warning("AuthenticationServer.get_pid not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_name(self, *args):
		logger.warning("AuthenticationServer.get_name not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def login_with_context(self, *args):
		logger.warning("AuthenticationServer.login_with_context not implemented")
		raise common.RMCError("Core::NotImplemented")


class AuthenticationServerNX(AuthenticationProtocolNX):
	def __init__(self):
		self.methods = {
			self.METHOD_VALIDATE_AND_REQUEST_TICKET: self.handle_validate_and_request_ticket,
			self.METHOD_VALIDATE_AND_REQUEST_TICKET_WITH_CUSTOM_DATA: self.handle_validate_and_request_ticket_with_custom_data,
			self.METHOD_REQUEST_TICKET: self.handle_request_ticket,
			self.METHOD_GET_PID: self.handle_get_pid,
			self.METHOD_GET_NAME: self.handle_get_name,
			self.METHOD_VALIDATE_AND_REQUEST_TICKET_WITH_PARAM: self.handle_validate_and_request_ticket_with_param,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on AuthenticationServerNX: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_validate_and_request_ticket(self, client, input, output):
		logger.info("AuthenticationServerNX.validate_and_request_ticket()")
		#--- request ---
		username = input.string()
		response = await self.validate_and_request_ticket(client, username)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'pid', 'ticket', 'connection_data', 'server_name']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.pid(response.pid)
		output.buffer(response.ticket)
		output.add(response.connection_data)
		output.string(response.server_name)
	
	async def handle_validate_and_request_ticket_with_custom_data(self, client, input, output):
		logger.info("AuthenticationServerNX.validate_and_request_ticket_with_custom_data()")
		#--- request ---
		username = input.string()
		extra_data = input.anydata()
		response = await self.validate_and_request_ticket_with_custom_data(client, username, extra_data)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'pid', 'ticket', 'connection_data', 'server_name', 'source_key']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.pid(response.pid)
		output.buffer(response.ticket)
		output.add(response.connection_data)
		output.string(response.server_name)
		output.string(response.source_key)
	
	async def handle_request_ticket(self, client, input, output):
		logger.info("AuthenticationServerNX.request_ticket()")
		#--- request ---
		source = input.pid()
		target = input.pid()
		response = await self.request_ticket(client, source, target)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'ticket', 'key']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.buffer(response.ticket)
		output.string(response.key)
	
	async def handle_get_pid(self, client, input, output):
		logger.info("AuthenticationServerNX.get_pid()")
		#--- request ---
		username = input.string()
		response = await self.get_pid(client, username)
		
		#--- response ---
		if not isinstance(response, int):
			raise RuntimeError("Expected int, got %s" %response.__class__.__name__)
		output.pid(response)
	
	async def handle_get_name(self, client, input, output):
		logger.info("AuthenticationServerNX.get_name()")
		#--- request ---
		pid = input.pid()
		response = await self.get_name(client, pid)
		
		#--- response ---
		if not isinstance(response, str):
			raise RuntimeError("Expected str, got %s" %response.__class__.__name__)
		output.string(response)
	
	async def handle_validate_and_request_ticket_with_param(self, client, input, output):
		logger.info("AuthenticationServerNX.validate_and_request_ticket_with_param()")
		#--- request ---
		param = input.extract(ValidateAndRequestTicketParam)
		response = await self.validate_and_request_ticket_with_param(client, param)
		
		#--- response ---
		if not isinstance(response, ValidateAndRequestTicketResult):
			raise RuntimeError("Expected ValidateAndRequestTicketResult, got %s" %response.__class__.__name__)
		output.add(response)
	
	async def validate_and_request_ticket(self, *args):
		logger.warning("AuthenticationServerNX.validate_and_request_ticket not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def validate_and_request_ticket_with_custom_data(self, *args):
		logger.warning("AuthenticationServerNX.validate_and_request_ticket_with_custom_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def request_ticket(self, *args):
		logger.warning("AuthenticationServerNX.request_ticket not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_pid(self, *args):
		logger.warning("AuthenticationServerNX.get_pid not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def get_name(self, *args):
		logger.warning("AuthenticationServerNX.get_name not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def validate_and_request_ticket_with_param(self, *args):
		logger.warning("AuthenticationServerNX.validate_and_request_ticket_with_param not implemented")
		raise common.RMCError("Core::NotImplemented")





from Crypto.Hash import CMAC
from Crypto.Cipher import AES
from anynet import tls, http
from nintendo import resources
import base64
import json
import time

import logging
logger = logging.getLogger(__name__)


DAUTH_SOURCE = bytes.fromhex("8be45abcf987021523ca4f5e2300dbf0")

SYSTEM_VERSION_DIGEST = {
	900:  "CusHY#00090000#-80vwBkUjWLb5Kpb_cnuTjBZ0rHwZHhN7R1-vg0Ti5c=",
	901:  "CusHY#00090001#qVDSOCehwMDCHyDnkXiTSJ1wEJZHtpRV_CLMKgD-fSw=",
	910:  "CusHY#00090100#vIPNrRbf30SoU8ZJ6uGklMqKAkyjHfdE9m6yLFeChkE=",
	920:  "CusHY#00090200#Uxxmc8gYnfMqxzdZdygZ_OrKo98O7QA65s_EkZnGsDo=",
	1000: "CusHY#000a0000#EmdxOnZjZ9Ihf3Zskt_48pYgowAUeeJccU6tCBIweEc=",
	1001: "CusHY#000a0001#JEuSEdid24qqHqQzfW1tuNsCGcCk-86gcPq0I7M1x18=",
	1002: "CusHY#000a0002#BTOGo0giC7bbrNoi8JEm-FBzmXb2Kgpq4K3OzQrD5l8=",
	1003: "CusHY#000a0003#4mBbTFYnE0Rtmh8NLCVq61rbvx0kJPQUxXkDpwj0V84=",
	1004: "CusHY#000a0003#4mBbTFYnE0Rtmh8NLCVq61rbvx0kJPQUxXkDpwj0V84=",
	1010: "CusHY#000a0100#Vlw9dIEqjxE2F5jDOQPYWXs2p7wIGyDYWXXIyQGdxcE=",
	1011: "CusHY#000a0100#Vlw9dIEqjxE2F5jDOQPYWXs2p7wIGyDYWXXIyQGdxcE=",
	1020: "CusHY#000a0200#90k0dE_eO7hRcs6ByTZMvgUm4lhEoqAlik96WkznQcQ=",
	1100: "CusHY#000b0000#VyA0fsWi6ZBEOzVsseXIcEfFLqQMgW0tWzN2oJ7viqk=",
	1101: "CusHY#000b0001#iI0rZ0Q2dg3Evhd-8GoYmp-KTE8malKe0GOJgXa-XG8=",
	1200: "CusHY#000c0000#C-BynYNPXdQJNBZjx02Hizi8lRUSIKLwPGa5p8EY1uo=",
	1201: "CusHY#000c0001#YXsU5FTbkUh18QH27L3woGqw5n1gIDpMGbPXM8oACzY=",
	1202: "CusHY#000c0002#6tB3UVnmvT_nsNBMPSD-K1oe0IA1cYvYDyqDCjy2W_I=",
	1203: "CusHY#000c0003#E8Ph6vISWsJtN0E3hsfVRoZMG1Qqkc-qGRlAp-Bs2SI=",
	1210: "CusHY#000c0100#Hzs8Gtp6yKgGKMb732-5Q-NvbQcHCgBh_LQrrpg0bIs=",
	1300: "CusHY#000d0000#r1xneESd4PiTRYIhVIl0bK1ST5L5BUmv_uGPLqc4PPo=",
	1310: "CusHY#000d0100#plps6S3C43QHhkI2oNvYIFjNxQjTcLdUX2_biEI5w2w=",
	1320: "CusHY#000d0200#JFVNVuG9x3V5tRshdD9FdJjgHOmzsrgXHocEPvW5eMM=",
	1321: "CusHY#000d0201#V1i7M7oEhkDaH1lcGlHhGAnyHONMAnTAA6pGdZ7MFqc=",
	1400: "CusHY#000e0000#35hWb15SBXTnbUfTMLBz9sCnfheuRGis0OTZqa7l8yw=",
	1410: "CusHY#000e0100#ctIxSPR4jenzQNGc6y4zXIvzvF75ty53jN0T15Rjtmk=",
	1411: "CusHY#000e0101#uTt4IVydkYqwYArOFR3BzOCmw0MkEeF_tZxHENYDh4E=",
	1412: "CusHY#000e0102#jHk6_VwXVPPl3ijRZ5jRy5MIAcUW_Q2uFdfJ0vrjhCA=",
	1500: "CusHY#000f0000#MJE7we0zvVhAnXN9uCU7fQAM7GiqGHpR5ECuC9G_nuU=",
	1501: "CusHY#000f0001#TMqizZGvUaBZApmHHQE0Uo7vQ6xWuQxZ8JH87_HnnqI=",
	1600: "CusHY#00100000#k_VrW8iX7QgupPlYZYhg3dLEVDhqGN_iXW5Mm0VYEvQ=",
	1601: "CusHY#00100001#qHay53MkzVLOUU_Iy7_kyPlUMnaoi7HXCAmESYTft_c=",
	1602: "CusHY#00100002#qjeCnaxVt5NjGjxosJOMVw-ZyR219B3qgAB3YtSil6g=",
	1603: "CusHY#00100003#Lis4m_Z4pXlDAaBBxeRO66_glyu92IAf2-dHKNxYAJs=",
	1610: "CusHY#00100100#3kOLHmbuMWa-kHN-SMtCkgNvwdwBHIb2b4f60BNNmrw=",
	1700: "CusHY#00110000#ntMZ00Jmb7Rdu18Fy1FPZo7RO3h_MYIqxozbQQDcVaA=",
	1701: "CusHY#00110001#7CmXEDXEN8wnVu-e7WY6Cv5CvmzjuG6EnKEkf1_jaC8=",
	1800: "CusHY#00120000#U531L4Si9RbhOVeyVppe18WHkJ0k4_KzrNtygsekMNo=",
	1801: "CusHY#00120001#chuxR_O35JFyJq7dIlT8yP1A-j1yBcF-iU4iVDjHt9g=",
	1810: "CusHY#00120100#7pfwz-8raijuW2lv4UOi4Hukp-DuY898HEK6hEYUjSM=",
	1900: "CusHY#00130000#x2jf5al2EkqJmdvmnTFaL6s4ic7X68N0dY9jnwwcL98=",
	1901: "CusHY#00130001#6I1eeoSqDdjA7eXPDrsWBbdM-VxVhveQYiG1oNfNSt0=",
}

SYSTEM_VERSION_HEX = {
	2000: "7147e1386c9b6c15d8f14e6ed68c4b9a7f28fb9b",
	2001: "0b2540e5cd7498dd61f6caeca5136c73d9b1d21a",
	2010: "fa9b24a1d97b9adf5fe462f7f0ee97e6ed6294d0",
	2011: "9ffad64d79dd150490201461bdf66c8db963f57d",
	2015: "0605c36a7aa2535fb8989a0d133a0b96b0d97a12",
	2020: "ad939aa8ed828aa87f8ca9c6231bc76697298b8d",
	2030: "c6061464f68281807f4780ae737392516d4813c7",
	2040: "cc744ded0c0eb7b0a71917a97ec00926427cd652",
	2050: "c7f472f9b64b7b0cba46a25e6a610f70613a4792",
	2100: "f6b2425b6888a66590db104fc734891696e0ecb3",
	2101: "066a75e6fab7316de34b88b60d229a0b2729e421",
	2110: "0d7a6334ddc3d637f69ec3976b17a260efd68fd4",
	2120: "ff8d6ddacae7c7fd1287e22c3c88bb961acb290c",
}

USER_AGENT = {
	900:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	901:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	910:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	920:  "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 9.3.0.0)",
	1000: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1001: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1002: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1003: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1004: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1010: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1011: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1020: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 10.4.0.0)",
	1100: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 11.4.0.0)",
	1101: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 11.4.0.0)",
	1200: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1201: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1202: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1203: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1210: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 12.3.0.0)",
	1300: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.3.0.0)",
	1310: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.4.0.0)",
	1320: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.4.0.0)",
	1321: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 13.4.0.0)",
	1400: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
	1410: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
	1411: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
	1412: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 14.3.0.0)",
	1500: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 15.3.0.0)",
	1501: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 15.3.0.0)",
	1600: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 16.2.0.0)",
	1601: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 16.2.0.0)",
	1602: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 16.2.0.0)",
	1603: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 16.2.0.0)",
	1610: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 16.2.0.0)",
	1700: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 17.5.0.0)",
	1701: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 17.5.0.0)",
	1800: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 18.3.0.0)",
	1801: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 18.3.0.0)",
	1810: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 18.3.0.0)",
	1900: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 19.3.0.0)",
	1901: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 19.3.0.0)",
	2000: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2001: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2010: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2011: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2015: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2020: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2030: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2040: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2050: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 20.5.4.0)",
	2100: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
	2101: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
	2110: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
	2120: "libcurl (nnDauth; 16f4553f-9eee-4e39-9b61-59bc7c99b7c8; SDK 21.4.0.0)",
}

KEY_GENERATION = {
	900:  10,
	901:  10,
	910:  11,
	920:  11,
	1000: 11,
	1001: 11,
	1002: 11,
	1003: 11,
	1004: 11,
	1010: 11,
	1011: 11,
	1020: 11,
	1100: 11,
	1101: 11,
	1200: 11,
	1201: 11,
	1202: 11,
	1203: 11,
	1210: 11,
	1300: 13,
	1310: 13,
	1320: 13,
	1321: 13,
	1400: 14,
	1410: 14,
	1411: 14,
	1412: 14,
	1500: 15,
	1501: 15,
	1600: 16,
	1601: 16,
	1602: 16,
	1603: 16,
	1610: 16,
	1700: 17,
	1701: 17,
	1800: 17,
	1801: 17,
	1810: 17,
	1900: 19,
	1901: 19,
	2000: 20,
	2001: 20,
	2010: 20,
	2011: 20,
	2015: 20,
	2020: 20,
	2030: 20,
	2040: 20,
	2050: 20,
	2100: 21,
	2101: 21,
	2110: 21,
	2120: 21,
}

API_VERSION = {
	900: 6,
	901: 6,
	910: 6,
	920: 6,
	1000: 6,
	1001: 6,
	1002: 6,
	1003: 6,
	1004: 6,
	1010: 6,
	1011: 6,
	1020: 6,
	1100: 6,
	1101: 6,
	1200: 6,
	1201: 6,
	1202: 6,
	1203: 6,
	1210: 6,
	1300: 7,
	1310: 7,
	1320: 7,
	1321: 7,
	1400: 7,
	1410: 7,
	1411: 7,
	1412: 7,
	1500: 7,
	1501: 7,
	1600: 7,
	1601: 7,
	1602: 7,
	1603: 7,
	1610: 7,
	1700: 7,
	1701: 7,
	1800: 7,
	1801: 7,
	1810: 7,
	1900: 7,
	1901: 7,
	2000: 8,
	2001: 8,
	2010: 8,
	2011: 8,
	2015: 8,
	2020: 8,
	2030: 8,
	2040: 8,
	2050: 8,
	2100: 8,
	2101: 8,
	2110: 8,
	2120: 8,
}

LATEST_VERSION = 2120


CLIENT_ID_SCSI = 0x146C8AC7B8A0DB52
CLIENT_ID_ER = 0x16E96F76850156D1
CLIENT_ID_ATUM = 0x3117B250CAB38F45
CLIENT_ID_ESHOP = 0x41F4A6491028E3C4
CLIENT_ID_BCAT = 0x67BF9945B45248C6
CLIENT_ID_SATA = 0x6AC5A6873FE5F68C
CLIENT_ID_ACCOUNT_APPLET = 0x75FE236362FF5F8B
CLIENT_ID_ACCOUNT = 0x81333C548B2E876D
CLIENT_ID_NPNS = 0x83B72B05DC3278D7
CLIENT_ID_BAAS = 0x8F849B5D34778D8E
CLIENT_ID_BEACH = 0x93AF0ACB26258DE9
CLIENT_ID_SPROFILE = 0xBAD8156F44AC935A
CLIENT_ID_DRAGONS = 0xD5B6CAC2C1514C56
CLIENT_ID_SCSI_POLICY = 0xD98185ACB55994B4
CLIENT_ID_PCTL = 0xDC656EA03B63CF68
CLIENT_ID_PREPO = 0xDF51C436BC01C437
CLIENT_ID_PENNE = 0xE58171FE439390CE


PRELOADED_DEVICE_TOKENS = [
	CLIENT_ID_BAAS,
	CLIENT_ID_PCTL,
	CLIENT_ID_BEACH,
	CLIENT_ID_PREPO,
	CLIENT_ID_ER,
	CLIENT_ID_PENNE,
	CLIENT_ID_ACCOUNT,
	CLIENT_ID_ACCOUNT_APPLET,
	CLIENT_ID_SCSI,
	CLIENT_ID_DRAGONS,
	CLIENT_ID_SPROFILE
]

PRELOADED_EDGE_TOKENS = [
	(CLIENT_ID_BEACH, "akamai"),
	(CLIENT_ID_BCAT, "akamai"),
	(CLIENT_ID_SCSI_POLICY, "akamai"),
	(CLIENT_ID_SCSI, "akamai"),
	(CLIENT_ID_ESHOP, "akamai"),
	(CLIENT_ID_ATUM, "akamai"),
	(CLIENT_ID_ATUM, "fastly")
]


class DAuthError(Exception):
	UNAUTHORIZED_DEVICE = 4
	SYSTEM_UPDATE_REQUIRED = 7
	BANNED_DEVICE = 8
	INTERNAL_SERVER_ERROR = 9
	GENERIC = 14
	CHALLENGE_EXPIRED = 15
	WRONG_MAC = 16
	BROKEN_DEVICE = 17
	
	def __init__(self, response, error):
		self.response = response
		self.code = int(error["code"])
		self.message = error["message"]
	
	def __str__(self):
		return self.message


class DAuthClient:
	def __init__(self, keys):
		self.keys = keys
		
		self.request_callback = http.request
		
		ca = resources.certificate("Nintendo_CA_G3.der")
		self.context = tls.TLSContext()
		self.context.set_authority(ca)
		
		self.host = "dauth-lp1.ndas.srv.nintendo.net"
		
		self.power_state = "FA"
		self.region = 1

		self.system_version = LATEST_VERSION
		self.user_agent = USER_AGENT[self.system_version]
		self.system_version_hash = SYSTEM_VERSION_HEX[self.system_version]
		self.key_generation = KEY_GENERATION[self.system_version]
		self.api_version = API_VERSION[self.system_version]
		
	def set_request_callback(self, callback): self.request_callback = callback
	
	def set_context(self, context): self.context = context
	def set_certificate(self, cert, key): self.context.set_certificate(cert, key)
	
	def set_power_state(self, state): self.power_state = state
	def set_platform_region(self, region): self.region = region
	
	def set_host(self, host): self.host = host
	
	def set_system_version(self, version):
		if version not in USER_AGENT:
			raise ValueError("Unknown system version: %i" %version)
		self.system_version = version
		self.user_agent = USER_AGENT[version]
		self.key_generation = KEY_GENERATION[version]
		self.api_version = API_VERSION[version]
		if self.api_version < 8:
			self.system_version_hash = SYSTEM_VERSION_DIGEST[version]
		else:
			self.system_version_hash = SYSTEM_VERSION_HEX[version]
		
	async def request(self, req):
		if self.system_version < 1800:
			req.headers["Host"] = self.host
			req.headers["User-Agent"] = self.user_agent
			req.headers["Accept"] = "*/*"
			req.headers["X-Nintendo-PowerState"] = self.power_state
			req.headers["Content-Length"] = 0
			req.headers["Content-Type"] = "application/x-www-form-urlencoded"
		elif self.system_version < 2000:
			req.headers["Host"] = self.host
			req.headers["Accept"] = "*/*"
			req.headers["Content-Type"] = "application/x-www-form-urlencoded"
			req.headers["X-Nintendo-PowerState"] = self.power_state
			req.headers["Content-Length"] = 0
		else:
			req.headers["Host"] = self.host
			req.headers["Accept"] = "*/*"
			req.headers["User-Agent"] = self.user_agent
			if req.json:
				req.headers["Content-Type"] = "application/json"
			else:
				req.headers["Content-Type"] = "application/x-www-form-urlencoded"
			req.headers["X-Nintendo-PowerState"] = self.power_state
			req.headers["Content-Length"] = 0
		
		response = await self.request_callback(self.host, req, self.context)
		if response.json and "errors" in response.json:
			logger.error("DAuth server returned errors:")
			for error in response.json["errors"]:
				logger.error("  (%s) %s", error["code"], error["message"])
			raise DAuthError(response, response.json["errors"][0])
		response.raise_if_error()
		return response
		
	async def challenge(self):
		req = http.HTTPRequest.post("/v%i/challenge" %self.api_version)
		req.form = {
			"key_generation": self.key_generation
		}
		
		response = await self.request(req)
		return response.json
	
	async def request_token(self, client_id, vendor_id="akamai", *, edge_token):
		# This is a generic method to reduce code duplication between device_token and edge_token

		# Use one of the preload_* functions on system version 20.0.0 and later, to mimic the
		# behavior of a real Switch.
		if self.system_version >= 2000:
			raise ValueError("This method is only available up to system version 19.0.1")
		
		challenge = await self.challenge()
		data = base64.b64decode(challenge["data"] + "==", "-_")

		path = "/v%i/device_auth_token" %self.api_version
		if edge_token:
			path = "/v%i/edge_token" %self.api_version
		
		req = http.HTTPRequest.post(path)
		req.rawform = {
			"challenge": challenge["challenge"],
			"client_id": "%016x" %client_id,
			"ist": "true" if self.region == 2 else "false",
			"key_generation": self.key_generation,
			"system_version": self.system_version_hash
		}
		if self.api_version >= 7 and edge_token:
			req.rawform["vendor_id"] = vendor_id

		string = http.formencode(req.rawform, False)
		req.rawform["mac"] = self.calculate_mac(string, data)

		response = await self.request(req)
		return response.json
	
	async def request_tokens(self, token_requests, *, edge_tokens):
		# This is a generic method to reduce code duplication between device_tokens and edge_tokens

		if self.system_version < 2000:
			raise ValueError("This method is only available on system version 20.0.0 and above.")
		
		challenge = await self.challenge()
		data = base64.b64decode(challenge["data"] + "==", "-_")

		system_version = "00%02x%02x%02x" %(
			self.system_version // 100,
			(self.system_version // 10) % 10,
			self.system_version % 10
		)

		form = {
			"challenge": challenge["challenge"],
			"fw_revision": self.system_version_hash,
			"ist": "true" if self.region == 2 else "false",
			"key_generation": self.key_generation,
			"system_version": system_version,
			"token_requests": json.dumps(token_requests, separators=(",", ":"))
		}
		string = http.formencode(form, False)
		mac = self.calculate_mac(string, data)

		path = "/v%i/device_auth_tokens" %self.api_version
		if edge_tokens:
			path = "/v%i/edge_tokens" %self.api_version

		req = http.HTTPRequest.post(path)
		req.json = {
			"system_version": system_version,
			"fw_revision": self.system_version_hash,
			"ist": self.region == 2,
			"token_requests": token_requests,
			"key_generation": self.key_generation,
			"challenge": challenge["challenge"],
			"mac": mac
		}

		response = await self.request(req)

		# Check if any errors have occurred
		errors = []
		for result in response.json["results"]:
			if "error" in result:
				errors.append(result["error"])
		
		if errors:
			logger.error("DAuth server returned errors:")
			for error in errors:
				logger.error("  (%s) %s", error["code"], error["message"])
			raise DAuthError(response, errors[0])

		return response.json
	
	async def device_token(self, client_id):
		return await self.request_token(client_id, edge_token=False)
	
	async def edge_token(self, client_id, vendor_id="akamai"):
		return await self.request_token(client_id, vendor_id, edge_token=True)
		
	async def device_tokens(self, client_ids):
		token_requests = [{"client_id": "%016x" %client_id} for client_id in client_ids]
		return await self.request_tokens(token_requests, edge_tokens=False)

	async def edge_tokens(self, token_requests):
		token_requests = [{"client_id": "%016x" %client_id, "vendor_id": vendor_id} for client_id, vendor_id in token_requests]
		return await self.request_tokens(token_requests, edge_tokens=True)

	async def preload_device_tokens(self):
		return await self.device_tokens(PRELOADED_DEVICE_TOKENS)
	
	async def preload_edge_tokens(self):
		return await self.edge_tokens(PRELOADED_EDGE_TOKENS)
		
	def get_master_key(self):
		keygen = self.key_generation
		keyname = "master_key_%02x" %(keygen - 1)
		return self.keys[keyname]
	
	def decrypt_key(self, key, kek):
		aes = AES.new(kek, AES.MODE_ECB)
		return aes.decrypt(key)
		
	def calculate_mac(self, form, data):
		kek_source = self.keys["aes_kek_generation_source"]
		master_key = self.get_master_key()
		
		key = self.decrypt_key(kek_source, master_key)
		key = self.decrypt_key(DAUTH_SOURCE, key)
		key = self.decrypt_key(data, key)
		
		mac = CMAC.new(key, ciphermod=AES)
		mac.update(form.encode())
		return base64.b64encode(mac.digest(), b"-_").decode().rstrip("=")


class DAuthCache:
	def __init__(self, client, expiration=None):
		self.client = client
		self.expiration = expiration

		self.device_tokens = {}
		self.edge_tokens = {}
	
	async def device_token(self, client_id):
		now = time.time()
		if client_id in self.device_tokens and self.device_tokens[client_id][1] > now:
			return self.device_tokens[client_id][0]
		
		if client_id in PRELOADED_DEVICE_TOKENS:
			response = await self.client.preload_device_tokens()
		else:
			response = await self.client.device_tokens([client_id])
		
		for result in response["results"]:
			if self.expiration is not None:
				expiration = now + self.expiration
			else:
				expiration = now + result["expires_in"]
			self.device_tokens[int(result["client_id"], 16)] = (result["device_auth_token"], expiration)

		return self.device_tokens[client_id][0]

	async def edge_token(self, client_id, vendor_id="akamai"):
		now = time.time()
		key = (client_id, vendor_id)
		if key in self.edge_tokens and self.edge_tokens[key][1] > now:
			return self.edge_tokens[key][0]
		
		if key in PRELOADED_EDGE_TOKENS:
			response = await self.client.preload_edge_tokens()
		else:
			response = await self.client.edge_tokens([key])
		
		for result in response["results"]:
			result_key = (int(result["client_id"], 16), result["vendor_id"])
			if self.expiration is not None:
				expiration = now + self.expiration
			else:
				expiration = now + result["expires_in"]
			self.edge_tokens[result_key] = (result["dtoken"], expiration)
		
		return self.edge_tokens[key][0]




from nintendo import resources


class Settings:
	TRANSPORT_UDP = 0
	TRANSPORT_TCP = 1
	TRANSPORT_WEBSOCKET = 2

	COMPRESSION_NONE = 0
	COMPRESSION_ZLIB = 1

	ENCRYPTION_NONE = 0
	ENCRYPTION_RC4 = 1
	
	field_types = {
		"nex.version": int,
		"nex.client_version": int,
		"nex.struct_header": int,
		"nex.pid_size": int,
		
		"prudp.access_key": str,
		
		"prudp.version": int,
		"prudp.minor_version": int,
		"prudp.supported_functions": int,
		
		"prudp.transport": int,
		"prudp.compression": int,
		"prudp.encryption": int,
		
		"prudp.resend_timeout": float,
		"prudp.resend_limit": int,
		"prudp.ping_timeout": float,
		
		"prudp.fragment_size": int,
		"prudp.max_substream_id": int,
		
		"prudp_v0.signature_version": int,
		"prudp_v0.flags_version": int,
		"prudp_v0.checksum_version": int,

		"kerberos.key_size": int,
		"kerberos.key_derivation": int,
		"kerberos.ticket_version": int
	}
	
	def __init__(self, filename=None):
		self.settings = {}
		self.reset()
		if filename:
			self.load(filename)
		
	def __getitem__(self, name): return self.settings[name]
	def __setitem__(self, name, value):
		if name not in self.field_types:
			raise KeyError("Unknown setting: %s" %name)
		self.settings[name] = self.field_types[name](value)
	
	def configure(self, access_key, nex_version, client_version=None):
		self["prudp.access_key"] = access_key
		self["nex.version"] = nex_version
		if nex_version >= 40400:
			if client_version is None:
				raise ValueError("NEX 4.4.0 or later requires client version")
			self["nex.client_version"] = client_version
		
	def reset(self): self.load("default")
	def copy(self):
		copy = Settings()
		copy.settings = self.settings.copy()
		return copy

	def load(self, name):
		with resources.open("files/config/%s.cfg" %name) as f:
			linenum = 1
			for line in f:
				line = line.strip()
				if line:
					if "=" in line:
						field, value = line.split("=", 1)
						self[field.strip()] = value.strip()
					else:
						raise ValueError("Syntax error at line %i" %linenum)
				linenum += 1
				
				
def default():
	return Settings()

def load(name):
	return Settings(name)




# This file was generated automatically by generate_protocols.py

from nintendo.nex import notification, rmc, common, streams

import logging
logger = logging.getLogger(__name__)


class ConnectionData(common.Structure):
	def __init__(self):
		super().__init__()
		self.station = None
		self.connection_id = None
	
	def check_required(self, settings, version):
		for field in ['station', 'connection_id']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.station = stream.stationurl()
		self.connection_id = stream.u32()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.stationurl(self.station)
		stream.u32(self.connection_id)


class NintendoLoginData(common.Structure):
	def __init__(self):
		super().__init__()
		self.token = None
	
	def check_required(self, settings, version):
		for field in ['token']:
			if getattr(self, field) is None:
				raise ValueError("No value assigned to required field: %s" %field)
	
	def load(self, stream, version):
		self.token = stream.string()
	
	def save(self, stream, version):
		self.check_required(stream.settings, version)
		stream.string(self.token)


class SecureConnectionProtocol:
	METHOD_REGISTER = 1
	METHOD_REQUEST_CONNECTION_DATA = 2
	METHOD_REQUEST_URLS = 3
	METHOD_REGISTER_EX = 4
	METHOD_TEST_CONNECTIVITY = 5
	METHOD_UPDATE_URLS = 6
	METHOD_REPLACE_URL = 7
	METHOD_SEND_REPORT = 8
	
	PROTOCOL_ID = 0xB


class SecureConnectionClient(SecureConnectionProtocol):
	def __init__(self, client):
		self.settings = client.settings
		self.client = client
	
	async def register(self, urls):
		logger.info("SecureConnectionClient.register()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(urls, stream.stationurl)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REGISTER, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.connection_id = stream.u32()
		obj.public_station = stream.stationurl()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.register -> done")
		return obj
	
	async def request_connection_data(self, cid, pid):
		logger.info("SecureConnectionClient.request_connection_data()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(cid)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REQUEST_CONNECTION_DATA, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.bool()
		obj.connection_data = stream.list(ConnectionData)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.request_connection_data -> done")
		return obj
	
	async def request_urls(self, cid, pid):
		logger.info("SecureConnectionClient.request_urls()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(cid)
		stream.pid(pid)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REQUEST_URLS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.bool()
		obj.urls = stream.list(stream.stationurl)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.request_urls -> done")
		return obj
	
	async def register_ex(self, urls, login_data):
		logger.info("SecureConnectionClient.register_ex()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(urls, stream.stationurl)
		stream.anydata(login_data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REGISTER_EX, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		obj = rmc.RMCResponse()
		obj.result = stream.result()
		obj.connection_id = stream.u32()
		obj.public_station = stream.stationurl()
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.register_ex -> done")
		return obj
	
	async def test_connectivity(self):
		logger.info("SecureConnectionClient.test_connectivity()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_TEST_CONNECTIVITY, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.test_connectivity -> done")
	
	async def update_urls(self, urls):
		logger.info("SecureConnectionClient.update_urls()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.list(urls, stream.stationurl)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_UPDATE_URLS, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.update_urls -> done")
	
	async def replace_url(self, url, new):
		logger.info("SecureConnectionClient.replace_url()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.stationurl(url)
		stream.stationurl(new)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_REPLACE_URL, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.replace_url -> done")
	
	async def send_report(self, report_id, data):
		logger.info("SecureConnectionClient.send_report()")
		#--- request ---
		stream = streams.StreamOut(self.settings)
		stream.u32(report_id)
		stream.qbuffer(data)
		data = await self.client.request(self.PROTOCOL_ID, self.METHOD_SEND_REPORT, stream.get())
		
		#--- response ---
		stream = streams.StreamIn(data, self.settings)
		if not stream.eof():
			raise ValueError("Response is bigger than expected (got %i bytes, but only %i were read)" %(stream.size(), stream.tell()))
		logger.info("SecureConnectionClient.send_report -> done")


class SecureConnectionServer(SecureConnectionProtocol):
	def __init__(self):
		self.methods = {
			self.METHOD_REGISTER: self.handle_register,
			self.METHOD_REQUEST_CONNECTION_DATA: self.handle_request_connection_data,
			self.METHOD_REQUEST_URLS: self.handle_request_urls,
			self.METHOD_REGISTER_EX: self.handle_register_ex,
			self.METHOD_TEST_CONNECTIVITY: self.handle_test_connectivity,
			self.METHOD_UPDATE_URLS: self.handle_update_urls,
			self.METHOD_REPLACE_URL: self.handle_replace_url,
			self.METHOD_SEND_REPORT: self.handle_send_report,
		}
	
	async def logout(self, client):
		pass
	
	async def handle(self, client, method_id, input, output):
		if method_id in self.methods:
			await self.methods[method_id](client, input, output)
		else:
			logger.warning("Unknown method called on SecureConnectionServer: %i", method_id)
			raise common.RMCError("Core::NotImplemented")
	
	async def handle_register(self, client, input, output):
		logger.info("SecureConnectionServer.register()")
		#--- request ---
		urls = input.list(input.stationurl)
		response = await self.register(client, urls)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'connection_id', 'public_station']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.u32(response.connection_id)
		output.stationurl(response.public_station)
	
	async def handle_request_connection_data(self, client, input, output):
		logger.info("SecureConnectionServer.request_connection_data()")
		#--- request ---
		cid = input.u32()
		pid = input.pid()
		response = await self.request_connection_data(client, cid, pid)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'connection_data']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.bool(response.result)
		output.list(response.connection_data, output.add)
	
	async def handle_request_urls(self, client, input, output):
		logger.info("SecureConnectionServer.request_urls()")
		#--- request ---
		cid = input.u32()
		pid = input.pid()
		response = await self.request_urls(client, cid, pid)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'urls']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.bool(response.result)
		output.list(response.urls, output.stationurl)
	
	async def handle_register_ex(self, client, input, output):
		logger.info("SecureConnectionServer.register_ex()")
		#--- request ---
		urls = input.list(input.stationurl)
		login_data = input.anydata()
		response = await self.register_ex(client, urls, login_data)
		
		#--- response ---
		if not isinstance(response, rmc.RMCResponse):
			raise RuntimeError("Expected RMCResponse, got %s" %response.__class__.__name__)
		for field in ['result', 'connection_id', 'public_station']:
			if not hasattr(response, field):
				raise RuntimeError("Missing field in RMCResponse: %s" %field)
		output.result(response.result)
		output.u32(response.connection_id)
		output.stationurl(response.public_station)
	
	async def handle_test_connectivity(self, client, input, output):
		logger.info("SecureConnectionServer.test_connectivity()")
		#--- request ---
		await self.test_connectivity(client)
	
	async def handle_update_urls(self, client, input, output):
		logger.info("SecureConnectionServer.update_urls()")
		#--- request ---
		urls = input.list(input.stationurl)
		await self.update_urls(client, urls)
	
	async def handle_replace_url(self, client, input, output):
		logger.info("SecureConnectionServer.replace_url()")
		#--- request ---
		url = input.stationurl()
		new = input.stationurl()
		await self.replace_url(client, url, new)
	
	async def handle_send_report(self, client, input, output):
		logger.info("SecureConnectionServer.send_report()")
		#--- request ---
		report_id = input.u32()
		data = input.qbuffer()
		await self.send_report(client, report_id, data)
	
	async def register(self, *args):
		logger.warning("SecureConnectionServer.register not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def request_connection_data(self, *args):
		logger.warning("SecureConnectionServer.request_connection_data not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def request_urls(self, *args):
		logger.warning("SecureConnectionServer.request_urls not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def register_ex(self, *args):
		logger.warning("SecureConnectionServer.register_ex not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def test_connectivity(self, *args):
		logger.warning("SecureConnectionServer.test_connectivity not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def update_urls(self, *args):
		logger.warning("SecureConnectionServer.update_urls not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def replace_url(self, *args):
		logger.warning("SecureConnectionServer.replace_url not implemented")
		raise common.RMCError("Core::NotImplemented")
	
	async def send_report(self, *args):
		logger.warning("SecureConnectionServer.send_report not implemented")
		raise common.RMCError("Core::NotImplemented")

